//
//
//      TOLKIEN: Tool Kit for Genetics-based Applications
//
//      Anthony Yiu-Cheung Tang, 1992-93, 1996.
//
//      Description
//      -----------
//      This program uses the GENITOR genetic algorithm alphabetize
//      a sequence of characters
//
//      each chromosome is a character array of the 26 characters 'a' - 'z'
//      shuffled in a random order
//      
//      execute the program with no arguments to see the options available
//
//      The idea of this program came from an example in
//      Matthew Wall's GAlib (obtainable from ftp://lancet.mit.edu/pub/ga/)
//

#include <iostream.h>
#include <hpldbin.h>
#include <genitor.h>
#include <GetOpt.h>
#include <charhpld.h>
#include <chromo.h>
#include <xover.h>

unsigned nGen = 400 , uPopSize = 30;
float    flMRate = 1 / 26.0;
float    flXRate = 1;                              

//
// The fitness function returns the number of genes that are
// in the correct position
//
class Alphabetizer
{ 
public:
	double operator()(const TGAObj * obj) 
        {
           vector<char>::const_iterator i, end;
           int score = 0;
           char c = 'a';
           for (i = ((const Haploid<char> *) obj)->begin(),
                end = ((const Haploid<char> *) obj)->end();
                i != end;
                i++, c++) 
           {
                score += (*i == c);
           }

           return score;
        }
};

void    printBanner(const char *prog)
{
        cout << "This program evolves a bit string with all one's" << endl;
        cout << "usage: " << prog << " [-n gen] [-p popsize] [-x xp t] [-m mut]" << endl;
        cout << "-n gen      the number of generations to perform" << endl;
        cout << "-p popsize  the population size" << endl;
        cout << "-x xrate    the crossover rate" << endl;
        cout << "-m mrate    the mutation rate" << endl;
}

void    main(int argc, char *argv[])                   
{
        GetOpt options(argc,argv,"n:p:x:m:g:");
        char opt;
                    
        if (argc == 1) {   
           printBanner(argv[0]);                     
           exit(-1);
        }
        else {                                           
           opt = options();                             
           if (opt < 0 || opt == EOF) {
               printBanner(argv[0]);
               exit(-1);                                                       
           }
           else      
             do {
                 switch (opt) {
                 case 'n' :
                    nGen = abs(atoi(options.optarg));
                    break;
                 case 'p' :
                    uPopSize = atoi(options.optarg);
                    break;
                 case 'x' :
                    flXRate = fabs(atof(options.optarg));
                    if (flXRate > 1)
                        flXRate = 1;
                    break;
                 case 'm' :
                    flMRate = fabs(atof(options.optarg));
                    if (flMRate > 1)
                        flMRate = 1;
                    break;
                 case '?' :
                    printBanner(argv[0]);
                    exit(-1);
                 }
             } while ((opt = options()) != EOF);
        }

        cout << "Generations       : " << nGen << endl;
        cout << "Population size   : " << uPopSize << endl;
        cout << "Crossover rate    : " << flXRate << endl;
        cout << "Mutation rate     : " << flMRate << endl;

        GENITOR<Alphabetizer>  ga(new CharHaploid(), uPopSize, 
                                  new OrderCrossover< Haploid<char> >(flXRate),
                                  flMRate);
                              
        while (ga.trials() < nGen) 
             ga.evolve();

        cout << "Final population" << endl;
        cout << "----------------" << endl;
	ga.printOn(cout);
        cout << endl;
        cout << "Selection performed : " << ga.numSelected() << endl;
        cout << "Crossover performed : " << ga.numCrossed() << endl;
        cout << "Best Individual : " ;
        ga.bestInd()->printOn(cout);
        cout << endl;
        cout << "Worst Individual : " ;
        ga.worstInd()->printOn(cout);
        cout << endl;
        cout << "Online : " << ga.objOnLine() << endl;
        cout << "Offline : " << ga.objOffLine() << endl;
}

