//
//      TOLKIEN: Tool Kit for Genetics-based Applications
//
//      Anthony Yiu-Cheung Tang, 1992-93.
//      Department of Computer Science
//      Chinese University of Hong Kong
//
#if !defined (__MMFUNC_H)
#define  __MMFUNC_H

#include <hpldbin.h>
#include <math.h>

#ifdef M_PI
#define PI M_PI
#endif

#ifndef PI
#define PI 3.14159265358979323846
#endif

const   double FOUR_PI = 4 * PI;

//
//              0.0 <= x <= 10.0
//              use 20 bits
//
//              y = x / 2.5     if   0.0 <= x <= 2.5
//                = 2 - x / 2.5      2.5  < x <= 5.0
//                = (x - 5) / 2.5    5.0  < x <= 7.5
//                = 4 - x / 2.5      7.5  < x <= 10.0
//
//
ParameterMap params1(20,0,10);

class Twopeaks
{
public:
        double operator()(const TGAObj *ind) {
            double  flX = params1((BinIndividual*) ind , 0);

            if (flX > 7.5)
                return 4 - flX / 2.5;
            if (flX > 5)
                return (flX - 5) / 2.5;
            if (flX > 2.5)
                return 2 - flX / 2.5;
            return flX / 2.5;
        }
};

//
//      A modified version of Himmelblau's Function
//
//
//                      2           2           2     2                       
//      F(x1,x2) = - (x1  + x2 - 11)  - (x1 + x2  - 7)    + 200
//
//              -5.0 <= xi <= 5.0
//              use 20 bits for xi, therefore chromosome length is 40
//
//
ParameterMap params2(20,-5,5);
class Hmb
{
public:
     double  operator()(const TGAObj* obj)
     {
        double x1 = params2((BinIndividual *) obj, 0);
        double x2 = params2((BinIndividual *) obj, 1);

	return 200 - pow(x1 * x1 + x2 - 11, 2.0) - 
                     pow(x1 + x2 * x2 - 7, 2.0);
     }
};

//
//      A function of five equal maxima
//
//                6
//	f(x) = sin (5 * PI * x)
//
//	   0.0 <= x <= 1.0
//		use 30 bits for chromosome
//
ParameterMap params3(30,0,1);
class Evenpeaks
{
public:
     double operator()(const TGAObj* ind)
     {
        double  flX = params3((BinIndividual*) ind , 0);

	return pow(sin(5 * M_PI * flX), 6.0);
     }
};

//
//      Five Unevenly spaced, but equal maxima
//
//                6             3/4
//	f(x) = sin (5 * PI * ( x     - 0.05))
//
//	   0.0 <= x <= 1.0
//		use 30 bits for chromosome
//
//
class Unevenpeaks
{
public:
    double operator()(const TGAObj* ind)
    {
        double  flX = params3((BinIndividual*) ind , 0);

	return pow(sin(5 * M_PI * (pow(flX, 0.75) - 0.05)), 6.0);
    }
};

//
//      A function with five uneven decreasing maxima
//
//                                                   2        6             3/4
//	f(x) = exp(-2 * log(2) * ((x - 0.08) / 0.854)  ) * sin (5 * PI * ( x    - 0.05))
//
//	   0.0 <= x <= 1.0
//		use 30 bits for chromosome
//
class Unevendecpeaks
{
public:
    double operator()(const TGAObj* ind)
    {
        double  flX = params3((BinIndividual*) ind , 0);

	return exp(-2 * log(2) * pow((flX - 0.08)/0.854,2.0)) *
	       pow(sin(5 * M_PI * (pow(flX, 0.75) - 0.05)), 6.0);
    }
};

//
// 	f(x) = 160/10 * x if x < 10
//	     = 200/5 * (x - 15) if x > 15
//	     = 160/5 * (15 - x) otherwise
//
//	    0.0 <= x <= 20.0
//		use 20 bits for chromosome
//
ParameterMap params4(20,0,20);

class Twopeaktrap
{
public:
    double operator()(const TGAObj* ind)
    {
        double  flX = params4((BinIndividual*) ind, 0);
	if (flX < 10.0)
            return 16 * flX;
	else if (flX > 15.0)
               return 40 * (flX - 15);
	else
	    return 32 * (15 - flX);
    }
};

//
//                             2
//  	    f(x) = 1 + cos(4*M_PI*|x|  )
//
//	    0.0 <= x <= 1.0
//		use 30 bits for chromosome
//
class Unevencos
{
public:
    double operator()(const TGAObj* ind)
    {
        double  flX = params3((BinIndividual*) ind , 0);
	return 1 + cos(FOUR_PI * pow(fabs(flX),2.0));
    }
};

class	SinInd : public BinHaploid
//
// 	thirty bits to represent values in [0.0, 1.0]
//
{
public:
	SinInd() : BinHaploid(30,true,true) {}
	SinInd(const SinInd& src) : BinHaploid(src)
	{
	}

        SinInd* clone() const { return new SinInd(*this); }

        void printOn(ostream& out) const
	{
		out << asDouble() * params3.precision() << ':' << objValue();
	}
};

class   HmbInd : public BinHaploid
{
public :
	HmbInd() : BinHaploid(40,true,true) { }

	HmbInd(const HmbInd& src) : BinHaploid(src)
	{
	}

        HmbInd* clone() const { return new HmbInd(*this); }

        void printOn(ostream& out) const
          {
                out << params2(this,0) << ',' << params2(this,1) << ':' << objValue();
          }
};

#endif

