// $Header: /home/anthony/tolkien/include/RCS/binind.h,v 1.8 1996/06/19 02:19:53 anthony Exp $
//
//      TOLKIEN: Tool Kit for Genetics-based Applications
//
//      Anthony Yiu-Cheung Tang, 1992-93.
//      Department of Computer Science
//      The Chinese University of Hong Kong.
//
//      Anthony Yiu-Cheung Tang, 1996.
//
#if !defined( __BININD_H )
#define     __BININD_H

#include <BitString.h>
#include <tgaobj.h>
#include <xover.h>

class BinHaploid;
class BinDiploid;

//
// The base class for binary chromosomes (diploids and haploids)
//
class BinIndividual : public TGAObj
{
public:

    virtual int  test(int) const = 0;
    virtual void invert(int) = 0;
    virtual void set(int) = 0;
    virtual void clear(int) = 0;
    virtual const BitString & asBitString() const = 0;

    // binary to numeric conversions
    virtual double	   asDouble(int,int) const = 0;
    virtual unsigned long  asLong(int,int) const = 0;
    virtual double	   asDouble() const = 0;
    virtual unsigned long  asLong() const = 0;

    virtual void operator=(const BinDiploid &) = 0;
    virtual void operator=(const BinHaploid &) = 0;
    virtual void operator=(const BitString &) = 0;

    // whether the bits are interpreted as Gray code
    virtual bool isGrayCode() const = 0;
    virtual void isGrayCode(bool flag) = 0;

    virtual unsigned hammingDistance(const BinIndividual &) const = 0;

    //
    // crossover functions
    //
    virtual TGAObj* onePtCrossover(const BinIndividual*) const = 0;
    virtual TGAObj* twoPtCrossover(const BinIndividual*) const = 0;
    virtual TGAObj*  oddPtCrossover(const TGAObj *, int) const = 0;
    virtual TGAObj*  evenPtCrossover(const TGAObj *, int) const = 0;
    virtual TGAObj*  uniformCrossover(const TGAObj *, double) const = 0;

    //
    // helper functions for crossover, no supposed to be called
    // by users but still have to be public
    //
    virtual TGAObj* _onePtCrossover(const BinDiploid *) const = 0;
    virtual TGAObj* _twoPtCrossover(const BinDiploid *) const = 0;
    virtual TGAObj* _oddPtCrossover(const BinHaploid *, int) const = 0;
    virtual TGAObj* _oddPtCrossover(const BinDiploid *, int) const = 0;
    virtual TGAObj* _evenPtCrossover(const BinHaploid *, int) const = 0;
    virtual TGAObj* _evenPtCrossover(const BinDiploid *, int) const = 0;
    virtual TGAObj* _uniformCrossover(const BinHaploid *, double) const = 0;
    virtual TGAObj* _uniformCrossover(const BinDiploid *, double) const = 0;
};


//
// one point crossover, the base case for oddPtCrossover
//
class BinOnePtCrossover : public PtCrossover
{
public:
      BinOnePtCrossover(double flRate=DEFAULT_XOVERATE) : PtCrossover()
      {
           nXpts = 1;
           rate(flRate);
      }

      TGAObj*  create()
      {
           const TGAObj* pTmp = pFather;
           TGAObj*  pChild;
           
           if (flip())
              pChild = ((const BinIndividual*) pMother)->
                       onePtCrossover((const BinIndividual*) pFather);
           else
              pChild = pMother->clone();

           // swap parents
           pFather = pMother;
           pMother = pTmp;

           count();

           return pChild;
      }

      Crossover* clone() const
      {
           return new BinOnePtCrossover(*this);
      }

      void xpts(unsigned) { }
};

//
// two points crossover, the base case for evenPtCrossover
//
class BinTwoPtCrossover : public PtCrossover
{
public:
      BinTwoPtCrossover(double flRate=DEFAULT_XOVERATE) : PtCrossover()
      {
           nXpts = 2;
           rate(flRate);
      }
      TGAObj*  create()                                            
      {
           const TGAObj* pTmp = pFather;                  
           TGAObj*  pChild;

           if (flip())
              pChild = ((const BinIndividual*) pMother)->
                       twoPtCrossover((const BinIndividual*) pFather);
           else
              pChild = pMother->clone();

           // swap parents
           pFather = pMother;                                          
           pMother = pTmp;

           count();
        
           return pChild;
      }

      Crossover* clone() const
      {
           return new BinTwoPtCrossover(*this);
      }

      void xpts(unsigned) { }
};

//
// Map a segment of bits in a binary string to a value in a certain range
// commonly used in function optimizations in which a binary string
// is used to encode the parameters for the function being optimized
//
class ParameterMap
{
public:
	ParameterMap(int uLen, double flMin, double flMax) : 
                 numBits(uLen), min(flMin)
        {
                precs = (flMax - flMin) / (pow(2.0, (double) numBits) - 1);
        }
        
        double operator()(const BinIndividual* ind, int i)
        {
                return ind->asDouble(numBits * i, numBits) * precs + min;
        }

        double precision() const { return precs; }

private:
	double precs;
	double min;
        int    numBits;
};

#endif

