// $Header: /home/anthony/tolkien/include/RCS/xover.h,v 1.5 1996/06/19 02:19:53 anthony Exp $
//
//      TOLKIEN: Tool Kit for Genetics-based Applications
//
//      Anthony Yiu-Cheung Tang, 1992-93.
//      Department of Computer Science
//      The Chinese University of Hong Kong.
//
#if !defined( __XOVER_H )
#define __XOVER_H

#include <iostream.h>
#include <tgaobj.h>
#include <xover.h>
#include <trand.h>
#include <vector.h>

class   Crossover
{
public:
	virtual ~Crossover() {}

        virtual double rate() const = 0;
        virtual void  rate(double) = 0;
        virtual void  setMates(const TGAObj *, const TGAObj *) = 0;
        virtual void  getMates(const TGAObj * &, const TGAObj * &) const = 0;
        virtual TGAObj*   create() = 0;
	virtual Crossover* clone() const = 0;
        virtual long total() const = 0;

protected:
        virtual bool flip() const
        {
	    return tRand.flip(rate());
        }

        virtual void count() = 0;
};

class   PtCrossoverBase : public Crossover
{
public:
        double rate() const
        {
	    return flXRate;
        }

        void rate(double flNewRate)
        {
	    flXRate = rangeCheck(flNewRate,0.0,1.0,DEFAULT_XOVERATE);
        }
        
        long total() const 
        { 
            return numCreated; 
        }

        void setMates(const TGAObj * ind1, const TGAObj * ind2)
        {
            pMother = ind1;
            pFather = ind2;
        }

        void getMates(const TGAObj * & pInd1, const TGAObj * & pInd2) const
        {
	    pInd1 = pMother;
	    pInd2 = pFather;
        }

protected:
        double flXRate;
        long numCreated;
	const TGAObj *  pMother;
	const TGAObj *  pFather;

        PtCrossoverBase() : numCreated(0), pFather(NULL), 
                            pMother(NULL), flXRate(0) { }

        void count() { numCreated++; }
};

//
//      In case when pMother is not a kind of pFather, e.g. (pMother is Diploid
//      and pFather is Haploid, function create will always return a child
//      of the type of pMother
//

class  PtCrossover : public PtCrossoverBase
{
public:
        unsigned  xpts() const
        {
	    return nXpts;
        }

        virtual void xpts(unsigned) = 0;

protected:
	int nXpts;
        
        PtCrossover() : nXpts(0) {}
};

template <class Chromosome>
class  OddPtCrossover : public PtCrossover
{
public:
       OddPtCrossover(unsigned nPoints=1,
                      double flRate=DEFAULT_XOVERATE)
       {
	   xpts(nPoints);
	   rate(flRate);
       }

       OddPtCrossover(const OddPtCrossover<Chromosome> & src)
       {
	   nXpts = src.nXpts;
           numCreated = 0;
	   rate(src.rate());
	   src.getMates(pFather, pMother);
       }

       Crossover* clone() const
       {
	    return new OddPtCrossover<Chromosome>(*this);
       }

       TGAObj*    create()
       {
            const TGAObj* pTmp = pFather;
	    TGAObj* pChild;

            if (flip())
                pChild = ((const Chromosome *) pMother)->oddPtCrossover(pFather, nXpts);
            else
	        pChild = pMother->clone();
	    // swap parents
	    pFather = pMother;
	    pMother = pTmp;

	    count(); 

	    return pChild;
       }

       void xpts(unsigned uNewPts)
       {
	    nXpts = uNewPts;
	    if (nXpts % 2 == 0)
	        nXpts--;
	    if (nXpts < 1)
	        nXpts = 1;
       }
};

template <class Chromosome>
class  EvenPtCrossover : public PtCrossover
{
public:
       EvenPtCrossover(int nPoints = 2,
		       double flRate=DEFAULT_XOVERATE)
       {
	   xpts(nPoints);
	   rate(flRate);
       }

       EvenPtCrossover(const EvenPtCrossover<Chromosome> & src) 
       {
	   nXpts = src.nXpts;
           numCreated = 0;
	   rate(src.rate());
	   src.getMates(pFather, pMother);
       }

       Crossover* clone() const
       {
	    return new EvenPtCrossover<Chromosome>(*this);
       }

       TGAObj*  create()
       {
            const TGAObj* pTmp = pFather;
	    TGAObj*  pChild;

            if (flip())
                pChild = ((const Chromosome *) pMother)->evenPtCrossover(pFather, nXpts);
            else
	        pChild = pMother->clone();
	    // swap parents
	    pFather = pMother;
	    pMother = pTmp;

	    count(); 

            return pChild;
       }

       void xpts(unsigned uNewPts)
       {
            nXpts = uNewPts;
            if (nXpts % 2 != 0)
	        nXpts--;
	    if (nXpts < 2)
	        nXpts = 2;
       }
};

template <class Chromosome>
class  MultiPtCrossover : public Crossover
{
public:
	MultiPtCrossover(unsigned uPoints = 2,
		         double flRate=DEFAULT_XOVERATE)
        {
            if (uPoints % 2 == 0)
                pXOver = new EvenPtCrossover<Chromosome>(uPoints, flRate);
            else
                pXOver = new OddPtCrossover<Chromosome>(uPoints, flRate);
        }

        MultiPtCrossover(const MultiPtCrossover<Chromosome> & src) 
        {
	    pXOver = (PtCrossover*) src.pXOver->clone();
        }

	~MultiPtCrossover()
        {
       	    delete pXOver;
        }

	unsigned  xpts() const
        {
	    return pXOver->xpts();
        }

        void xpts(unsigned uNewPts)
        {
	    unsigned nOldXpts = pXOver->xpts();
	    double flRate = pXOver->rate();
	    const TGAObj *pInd1, *pInd2;

            if (pXOver->xpts() % 2 == 0) {
                if (uNewPts % 2 == 0)
                    pXOver->xpts(uNewPts);
                else {
	            pXOver->getMates(pInd1, pInd2);
                    delete pXOver;
                    pXOver = new OddPtCrossover<Chromosome>(uNewPts, flRate);
		    pXOver->setMates(pInd1, pInd2);
                }
            }
            else {
                if (uNewPts % 2 != 0)
                    pXOver->xpts(uNewPts);
                else {
	            pXOver->getMates(pInd1, pInd2);
                    delete pXOver;
                    pXOver = new EvenPtCrossover<Chromosome>(uNewPts, flRate);
		    pXOver->setMates(pInd1, pInd2);
                }
            }
        } 

        Crossover* clone() const
        {
	    return pXOver->clone();
        }
        
        TGAObj*   create()
        {
	   return pXOver->create();
        }

        long total() const 
        {
            return pXOver->total();
        }

        void setMates(const TGAObj * mother, const TGAObj * father)
        {
	    pXOver->setMates(mother, father);
        }

        void getMates(const TGAObj * & pInd1, const TGAObj * & pInd2) const
        {
	    pXOver->getMates(pInd1, pInd2);
        }

        double rate() const
        {
	   return pXOver->rate();
        }

        void rate(double flNewRate)
        {
	   pXOver->rate(flNewRate);
        }

protected:
        PtCrossover* pXOver;

        void count() 
        { 
           // do nothing
        }
};

template <class Chromosome>
class   UniformCrossover : public Crossover
{
public:
	UniformCrossover(double flRate = DEFAULT_XOVERATE) : numCreated(0)
        {
       	    rate(flRate);
        }

	UniformCrossover(const UniformCrossover & src)
        {
       	    rate(src.rate());
            numCreated = 0;
	    src.getMates(pFather, pMother);
        }

        Crossover* clone() const
        {
	    return new UniformCrossover<Chromosome>(*this);
        }

	TGAObj*   create()
        {
            const TGAObj* pTmp = pFather;
	    TGAObj*  pChild = 
             ((const Chromosome*) pMother)->uniformCrossover(pFather, flXRate);
	    // swap parents
	    pFather = pMother;
	    pMother = pTmp;

	    count(); 

            return pChild;
        }

        void setMates(const TGAObj * mother, const TGAObj * father)
        {
            pMother = mother;
            pFather = father;
        }

        void getMates(const TGAObj * & pInd1, const TGAObj * & pInd2) const
        {
	    pInd1 = pMother;
	    pInd2 = pFather;
        }
    
        long total() const 
        { 
             return numCreated; 
        }

        double rate() const
        {
	     return flXRate;
        }

        void rate(double flNewRate)
        {
	     flXRate = rangeCheck(flNewRate,0.0,1.0,DEFAULT_XOVERATE);
        }

protected:
        double          flXRate;
	const TGAObj *  pMother;
	const TGAObj *  pFather;
        long numCreated;
        void count() { numCreated++; }
};

#endif

